import * as t from 'ts-codec';
export declare const BIGINT_MAX: bigint;
/**
 * The use of ts-codec:
 * We currently use pgwire for Postgres queries. This library provides fine-grained control
 * over parameter typings and efficient streaming of query responses. Additionally, configuring
 * pgwire with default certificates allows us to use the same connection configuration process
 * for both replication and storage libraries.
 *
 * Unfortunately, ORM driver support for pgwire is limited, so we rely on pure SQL queries in the
 * absence of writing an ORM driver from scratch.
 *
 * [Opinion]: Writing pure SQL queries throughout a codebase can be daunting from a maintenance
 * and debugging perspective. For example, row response types are often declared when performing a query:
 *
 * ```typescript
 * const rows = await db.queryRows<MyRowType>(`SELECT one, two FROM my_table`);
 * ```
 * This type declaration suggests `rows` is an array of `MyRowType` objects, even though no validation
 * is enforced. Adding a field to the `MyRowType` interface without updating the query could easily
 * introduce subtle bugs. Similarly, type mismatches between SQL results and TypeScript interfaces, such as
 * a `Date` field returned as a `string`, require manual conversion.
 *
 * `ts-codec` is not an ORM, but it simplifies working with pure SQL query responses in several ways:
 *
 * - **Validations**: The `decode` operation ensures that the returned row matches the expected object
 *   structure, throwing an error if it doesn't.
 * - **Decoding Columns**: pgwire already decodes common SQLite types, but `ts-codec` adds an extra layer
 *   for JS-native values. For instance, `jsonb` columns are returned as `JsonContainer`/`string` and can
 *   be automatically parsed into objects. Similarly, fields like `group_id` are converted from `Bigint`
 *   to `Number` for easier use.
 * - **Encoded Forms**: A single `ts-codec` type definition can infer both encoded and decoded forms. This
 *   is especially useful for persisted batch operations that rely on JSON query parameters for bulk inserts.
 *   Collections like `bucket_data`, `current_data`, and `bucket_parameters` use encoded/decoded types, making
 *   changes easier to manage and validate. While some manual encoding is done for intermediate values (e.g.,
 *   size estimation), these types are validated with `ts-codec` to ensure consistency.
 */
/**
 * Wraps a codec which is encoded to a JSON string
 */
export declare const jsonb: <Decoded>(subCodec: t.Codec<Decoded, any>) => t.Codec<Decoded, string, string, t.CodecProps>;
/**
 * Just performs a pure JSON.parse for the decoding step
 */
export declare const jsonb_raw: <Decoded>() => t.Codec<Decoded, string, string, t.CodecProps>;
export declare const bigint: t.Codec<bigint, string | number, string, t.CodecProps>;
export declare const uint8array: t.Codec<Uint8Array, Uint8Array, string, t.CodecProps>;
/**
 * PGWire returns BYTEA values as Uint8Array instances.
 * We also serialize to a hex string for bulk inserts.
 */
export declare const hexBuffer: t.Codec<Buffer, string, "hexBuffer", t.CodecProps>;
/**
 * PGWire returns INTEGER columns as a `bigint`.
 * This does a decode operation to `number`.
 */
export declare const pgwire_number: t.Codec<number, number, "pg_number", t.CodecProps>;
/**
 * A codec which contains the same type on the input and output.
 */
export declare const IdentityCodec: <T>() => t.Codec<T, T, string, t.CodecProps>;
